# pylint: disable=unsupported-membership-test,too-many-arguments
"""
Pyavatar Library
~~~~~~~~~~~~~~~~

Pyavatar is a library, written in Python, to generate simple default
user avatars to use in a web application or elsewhere.

:copyright: (c) 2020 by Matthieu Petiteau.
:license: MIT, see LICENSE for more details.
"""

import os
import random
from base64 import b64encode
from enum import Enum, IntEnum
from io import BytesIO
from typing import TypeAlias

from PIL import Image, ImageDraw, ImageFont

from ._version import __version__

__all__ = ("PyAvatar",
           "__version__",
           "PyAvatarError",
           "RenderingSizeError",
           "FontpathError",
           "FontExtensionNotSupportedError",
           "ImageExtensionNotSupportedError",
           )


class PyAvatarError(Exception):
    """Base PyAvatar error."""

    def __init__(self,
                 value: str,
                 message: str | None = None,
                 info: str = "") -> None:
        self.value = value
        self.message = message or self.__doc__
        self.info = info
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"{self.value} -> {self.message} {self.info}"


class RenderingSizeError(PyAvatarError):
    """Error with the chosen rendering size."""


class FontpathError(PyAvatarError):
    """Cannot find a font file at this location."""


class FontExtensionNotSupportedError(PyAvatarError):
    """Font file extension not supported."""


class ImageExtensionNotSupportedError(PyAvatarError):
    """Image extension not supported."""


class _BaseEnumUtils:

    @classmethod
    def get_set(cls) -> set[str]:
        return {attribute.value for attribute in cls}  # type: ignore # pylint: disable=not-an-iterable

    @classmethod
    def get_csv(cls) -> str:
        return ", ".join(cls.get_set())


class SupportedImageFormat(_BaseEnumUtils, str, Enum):
    PNG = "png"
    JPEG = "jpeg"
    ICO = "ico"


class SupportedFontFormat(_BaseEnumUtils, str, Enum):
    TTF = ".ttf"
    OTF = ".otf"


class SupportedPixelRange(_BaseEnumUtils, IntEnum):
    MIN = 50
    MAX = 650


_DEFAULT_IMAGE_SIZE = 120
_DEFAULT_FILEPATH = f"{os.getcwd()}/avatar.png"
_DEFAULT_FONT_FILEPATH = os.path.join(os.path.dirname(__file__),
                                      "font/Lora.ttf")

_HexColor: TypeAlias = str
_RGBColor: TypeAlias = tuple[int, int, int]


class PyAvatar:
    """Generate a default avatar from a given string input.

    :param text: Input text to use in the avatar.
    :param size: (optional) Integer, size in pixel of the avatar.
    :param fontpath: (optional) Filepath to the font file to use.
    :param color: (optional) hex or rgb color code for the background.
    :type color: string or tuple
    :param capitalize: (optional) Boolean, capitalize the first letter.
    :type capitalize: bool

    Usage::
      >>> from pyavatar import PyAvatar
      >>> avatar = PyAvatar("smallwat3r", size=250)
      >>> avatar.color
      (191, 91, 81)
      >>> avatar.change_color()
      >>> avatar.color
      (203, 22, 126)
      >>> avatar.stream("png")
      b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\xfa\x00\x00 ...'
      >>> avatar.base64_image("jpeg")
      'data:image/jpeg;base64,/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDAAgGBg ...'
      >>> import os
      >>> avatar.save(f"{os.getcwd()}/me.png")
    """

    def __init__(self,
                 text: str,
                 size: int = _DEFAULT_IMAGE_SIZE,
                 fontpath: str = _DEFAULT_FONT_FILEPATH,
                 color: _HexColor | _RGBColor | None = None,
                 capitalize: bool = True):
        self.text = text
        if capitalize:
            self.text = text.upper()
        self.size = size
        self.fontpath = fontpath
        self.color = color or self._random_color()
        self.image = self.__generate_avatar()

    def __str__(self) -> str:
        return f"{self.text} {self.size}x{self.size} {self.color}"

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("Attribute `text` must be a string.")
        self._text = value[0]  # keep the first letter

    @property
    def size(self) -> int:
        return self._size

    @size.setter
    def size(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError("Attribute `size` must be an integer.")
        if value < SupportedPixelRange.MIN or value > SupportedPixelRange.MAX:
            raise RenderingSizeError(
                str(value),
                f"Size must fit in range {SupportedPixelRange.get_set()}.")
        self._size = value

    @property
    def fontpath(self) -> str:
        return self._fontpath

    @fontpath.setter
    def fontpath(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError("Attribute `fontpath` must be a string.")
        if not os.path.exists(value):
            raise FontpathError(value)
        if not value.lower().endswith(
            (SupportedFontFormat.TTF, SupportedFontFormat.OTF)):
            raise FontExtensionNotSupportedError(
                os.path.basename(value),
                info=f"Supported formats: {SupportedFontFormat.get_csv()}.")
        self._fontpath = value

    @staticmethod
    def _random_color() -> _RGBColor:
        return (random.randint(0, 255),
                random.randint(0, 255),
                random.randint(0, 255))

    def __generate_avatar(self) -> Image:
        image = Image.new(mode="RGB",
                          size=(self.size, self.size),
                          color=self.color)
        font = ImageFont.truetype(self.fontpath, size=int(0.6 * self.size))
        draw = ImageDraw.Draw(image)
        w_txt, h_txt = draw.textsize(self.text, font)
        off_x, off_y = font.getoffset(self.text)
        position = ((self.size / 2 - (w_txt + off_x) / 2),
                    (self.size / 2 - (h_txt + off_y) / 2))
        draw.text(position, self.text, font=font)
        return image

    def change_color(self, color: _HexColor | _RGBColor | None = None) -> None:
        """Redraw the avatar with a new background color.

        :param color: (optional) hex or rgb color code for the background.
        :type color: string or tuple
        """
        self.color = color or self._random_color()
        self.image = self.__generate_avatar()

    def save(self, filepath: str = _DEFAULT_FILEPATH) -> None:
        """Save the avatar under a given file path.

        :param filepath: (optional) Filepath where the avatar will be saved.
        """
        extension = os.path.splitext(filepath)[1].split(".")[1]
        if extension not in SupportedImageFormat.get_set():
            raise ImageExtensionNotSupportedError(
                os.path.basename(filepath),
                info=f"Supported formats: {SupportedImageFormat.get_csv()}.")
        directory = os.path.dirname(filepath)
        if not os.path.exists(directory):
            os.makedirs(directory)
        self.image.save(filepath, optimize=True)

    def stream(
            self,
            filetype: SupportedImageFormat = SupportedImageFormat.PNG
    ) -> bytes:
        """Save the avatar in a bytes array.

        :param filetype: (optional) Avatar file format.
        :rtype: bytes
        """
        if filetype.lower() not in SupportedImageFormat.get_set():
            raise ImageExtensionNotSupportedError(
                filetype,
                info=f"Supported formats: {SupportedImageFormat.get_csv()}.")
        stream = BytesIO()
        self.image.save(stream, format=filetype, optimize=True)
        return stream.getvalue()

    def base64_image(
            self,
            filetype: SupportedImageFormat = SupportedImageFormat.PNG) -> str:
        """Save the avatar as a base64 image.

        :param filetype: (optional) Avatar file format.
        :rtype: str
        """
        encoded_image = b64encode(self.stream(filetype)).decode("utf-8")
        return f"data:image/{filetype};base64,{encoded_image}"



import os

from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt", "r") as f:
    install_requires = f.read().splitlines()

g = {}
with open(os.path.join("pyavatar", "_version.py"), "rt") as f:
    exec(f.read(), g)
    version = g["__version__"]

setup(name="pyavatar",
      version=version,
      author="Matthieu Petiteau",
      author_email="mpetiteau.pro@gmail.com",
      keywords="avatar",
      license="MIT",
      description="Generate simple user avatars from an input string.",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/smallwat3r/pyavatar",
      include_package_data=True,
      packages=find_packages(),
      install_requires=install_requires,
      classifiers=[
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      python_requires=">=3.10")



import os
import tempfile
import unittest

from pyavatar import (FontExtensionNotSupportedError,
                      FontpathError,
                      ImageExtensionNotSupportedError,
                      PyAvatar,
                      RenderingSizeError,
                      SupportedImageFormat,
                      SupportedPixelRange,
                      )


class TestPyAvatar(unittest.TestCase):
    """Pyavatar unittests."""

    name = "test"

    def test_avatar_creation(self):
        """Test avatar properties are correct."""
        avatar = PyAvatar(self.name, size=200, color=(9, 9, 9))
        self.assertEqual(avatar.text, "T")
        self.assertEqual(avatar.size, 200)
        self.assertEqual(avatar.color, (9, 9, 9))

    def test_avatar_text_capitalize(self):
        """Test capitalize text."""
        avatar = PyAvatar(self.name)
        self.assertEqual(avatar.text, self.name.upper()[0])
        avatar = PyAvatar(self.name, capitalize=False)
        self.assertEqual(avatar.text, self.name.lower()[0])

    def test_text_property_validation(self):
        """Test text property validation."""
        with self.assertRaises(TypeError):
            PyAvatar(123)
            PyAvatar(list())

    def test_size_property_validation(self):
        """Test size property validation."""
        with self.assertRaises(TypeError):
            PyAvatar(self.name, size="string")
            PyAvatar(self.name, size=list())
        with self.assertRaises(RenderingSizeError):
            PyAvatar(self.name, size=SupportedPixelRange.MAX + 1)
            PyAvatar(self.name, size=SupportedPixelRange.MIN - 1)

    def test_fontpath_property_validation(self):
        """Test fontpath property validation."""
        with self.assertRaises(TypeError):
            PyAvatar(self.name, fontpath=123)
            PyAvatar(self.name, fontpath=list())
        with self.assertRaises(FontpathError):
            PyAvatar(self.name, fontpath="idonotexist.ttf")
        with self.assertRaises(FontExtensionNotSupportedError):
            PyAvatar(self.name, fontpath=__file__)

    def test_change_color(self):
        """Test change avatar color."""
        avatar = PyAvatar(self.name)
        original_color = avatar.color
        avatar.change_color()
        self.assertNotEqual(original_color, avatar.color)

    def test_save_avatar(self):
        """Test save avatar locally."""
        avatar = PyAvatar(self.name)
        with tempfile.TemporaryDirectory() as temp_dir:
            filepath = f"{temp_dir}/new/test.png"
            avatar.save(filepath)
            self.assertTrue(os.path.isfile(filepath))
            with self.assertRaises(ImageExtensionNotSupportedError):
                filepath = f"{temp_dir}/test.nope"
                avatar.save(filepath)

    def test_stream_avatar(self):
        """Test save avatar in bytes array."""
        avatar = PyAvatar(self.name)
        stream = avatar.stream()
        self.assertIsInstance(stream, bytes)
        with self.assertRaises(ImageExtensionNotSupportedError):
            avatar.stream("unknown")

    def test_save_base64_avatar(self):
        """Test save avatar in base64."""
        avatar = PyAvatar(self.name)
        for format in SupportedImageFormat.get_set():
            image = avatar.base64_image(format)
            self.assertIsInstance(image, str)
            self.assertIn(format, image[:20])
